# import torch

# torch.backends.cuda.matmul.allow_tf32 = True


# import copy
# import warnings
# from datetime import timedelta

# from accelerate import Accelerator, DistributedType, InitProcessGroupKwargs
# from accelerate.state import AcceleratorState
# from lmms_eval import utils
# from lmms_eval.api.instance import Instance
# from lmms_eval.api.model import lmms
# from lmms_eval.api.registry import register_model
# from packaging import version
# from tqdm import tqdm

# warnings.filterwarnings("ignore")

# from loguru import logger as eval_logger

# try:
#     from vlm import VLM
#     from vlm.inference.generator import (
#         process_images,
#         tokenizer_image_token,
#     )
#     from vlm.utils.conversation import conv_templates
# except Exception as e:
#     eval_logger.debug(
#         "small-vlm is not installed. Please install small-vlm to use this model.\nError: %s" % e
#     )

# if version.parse(torch.__version__) >= version.parse("2.1.2"):
#     best_fit_attn_implementation = "sdpa"
# else:
#     best_fit_attn_implementation = "eager"


# @register_model("small-vlm")
# class SmallVLM(lmms):
#     """
#     Small VLM Model
#     """

#     def __init__(
#         self,
#         pretrained: str = "liuhaotian/llava-v1.5-7b",
#         truncation: bool | None = True,
#         device: str | None = "cuda:0",
#         batch_size: int | str | None = 1,
#         model_name=None,
#         attn_implementation=best_fit_attn_implementation,
#         device_map="cuda:0",
#         conv_template="vicuna_v1",
#         use_cache=True,
#         tie_weights: bool = True,
#         truncate_context=False,  # whether to truncate the context in generation, set it False for LLaVA-1.6
#         customized_config=None,  # ends in json
#         **kwargs,
#     ) -> None:
#         super().__init__()
#         # Do not use kwargs for now
#         assert kwargs == {}, f"Unexpected kwargs: {kwargs}"

#         accelerator_kwargs = InitProcessGroupKwargs(timeout=timedelta(weeks=52))
#         accelerator = Accelerator(kwargs_handlers=[accelerator_kwargs])
#         self.accelerator = accelerator
#         if accelerator.num_processes > 1:
#             self._device = torch.device(f"cuda:{accelerator.local_process_index}")
#             self.device_map = f"cuda:{accelerator.local_process_index}"
#         elif accelerator.num_processes == 1 and device_map == "auto":
#             self._device = torch.device(device)
#             self.device_map = device_map
#         else:
#             self._device = torch.device(f"cuda:{accelerator.local_process_index}")
#             self.device_map = f"cuda:{accelerator.local_process_index}"

#         small_vlm_model_args = {}
#         if customized_config is not None:
#             small_vlm_model_args["customized_config"] = customized_config
#         if attn_implementation is not None:
#             small_vlm_model_args["attn_implementation"] = attn_implementation
#         self._model = VLM.from_pretrained(pretrained, **small_vlm_model_args)
#         self._tokenizer = self.model.language_model.tokenizer
#         self._image_processor = self.model.visual_encoder.preprocessor
#         self._max_length = self.model.model_config.language_model.max_seq_length
#         self._config = self._model.config
#         self.image_token_index: int = self.model.model_config.language_model.image_token_index
#         self.image_token = self.model.model_config.language_model.image_token
#         self.model.eval()
#         if tie_weights:
#             self.model.tie_weights()

#         self.truncation = truncation
#         self.batch_size_per_gpu = int(batch_size)
#         self.conv_template = conv_template
#         self.use_cache = use_cache
#         self.truncate_context = truncate_context
#         # assert self.batch_size_per_gpu == 1, "Llava currently does not support batched generation. See https://github.com/haotian-liu/LLaVA/issues/754. HF Llava also has this issue."
#         if accelerator.num_processes > 1:
#             assert accelerator.distributed_type in [
#                 DistributedType.FSDP,
#                 DistributedType.MULTI_GPU,
#                 DistributedType.DEEPSPEED,
#             ], "Unsupported distributed type provided. Only DDP and FSDP are supported."
#             # If you want to use DistributedType.DEEPSPEED, you have to run accelerate config before using the model
#             # Also, you have to select zero stage 0 (equivalent to DDP) in order to make the prepare model works
#             # I tried to set different parameters in the kwargs to let default zero 2 stage works, but it didn't work.
#             if accelerator.distributed_type == DistributedType.DEEPSPEED:
#                 kwargs = {
#                     "train_micro_batch_size_per_gpu": self.batch_size_per_gpu,
#                     "train_batch_size": self.batch_size_per_gpu * accelerator.num_processes,
#                 }
#                 AcceleratorState().deepspeed_plugin.deepspeed_config_process(
#                     must_match=True, **kwargs
#                 )
#                 eval_logger.info(
#                     "Detected that you are using DistributedType.DEEPSPEED. Make sure you run `accelerate config` and set zero stage to 0"
#                 )

#             if (
#                 accelerator.distributed_type == DistributedType.FSDP
#                 or accelerator.distributed_type == DistributedType.DEEPSPEED
#             ):
#                 self._model = accelerator.prepare(self.model)
#             else:
#                 self._model = accelerator.prepare_model(self.model, evaluation_mode=True)
#             self.accelerator = accelerator
#             if self.accelerator.is_local_main_process:
#                 eval_logger.info(f"Using {accelerator.num_processes} devices with data parallelism")
#             self._rank = self.accelerator.local_process_index
#             self._world_size = self.accelerator.num_processes
#         elif accelerator.num_processes == 1 and device_map == "auto":
#             eval_logger.info(f"Using {accelerator.num_processes} devices with tensor parallelism")
#             self._rank = 0
#             self._world_size = 1
#         else:
#             eval_logger.info(f"Using single device: {self._device}")
#             self.model.to(self._device)
#             self._rank = 0
#             self._world_size = 1

#     @property
#     def config(self):
#         # return the associated transformers.AutoConfig for the given pretrained model.
#         return self._config

#     @property
#     def tokenizer(self):
#         return self._tokenizer

#     @property
#     def model(self):
#         # returns the model, unwrapping it if using Accelerate
#         if hasattr(self, "accelerator"):
#             return self.accelerator.unwrap_model(self._model)
#         else:
#             return self._model

#     @property
#     def eot_token_id(self):
#         # we use EOT because end of *text* is more accurate for what we're doing than end of *sentence*
#         return self.tokenizer.eos_token_id

#     @property
#     def max_length(self):
#         return self._max_length

#     def pad_sequence(self, input_ids, batch_first, padding_value):
#         if self.tokenizer.padding_side == "left":
#             input_ids = [torch.flip(_input_ids, [0]) for _input_ids in input_ids]
#         input_ids = torch.nn.utils.rnn.pad_sequence(
#             input_ids, batch_first=batch_first, padding_value=padding_value
#         )
#         if self.tokenizer.padding_side == "left":
#             input_ids = torch.flip(input_ids, [1])
#         return input_ids

#     @property
#     def batch_size(self):
#         return self.batch_size_per_gpu

#     @property
#     def device(self):
#         return self._device

#     @property
#     def rank(self):
#         return self._rank

#     @property
#     def world_size(self):
#         return self._world_size

#     def tok_encode(self, string: str, left_truncate_len=None, add_special_tokens=None) -> list[int]:
#         """ """
#         add_special_tokens = False if add_special_tokens is None else add_special_tokens
#         encoding = self.tokenizer.encode(string, add_special_tokens=add_special_tokens)
#         # left-truncate the encoded context to be at most `left_truncate_len` tokens long
#         if left_truncate_len:
#             encoding = encoding[-left_truncate_len:]
#         return encoding

#     def tok_decode(self, tokens):
#         try:
#             return self.tokenizer.decode(tokens)
#         except:
#             return self.tokenizer.decode([tokens])

#     def loglikelihood(self, requests: list[Instance]) -> list[tuple[float, bool]]:
#         # TODO
#         res = []
#         pbar = tqdm(total=len(requests), disable=(self.rank != 0), desc="Model Responding")

#         for contexts, doc_to_target, doc_to_visual, doc_id, task, split in [
#             reg.args for reg in requests
#         ]:
#             # encode, pad, and truncate contexts for this batch
#             if type(doc_to_target) == str:
#                 continuation = doc_to_target
#             else:
#                 continuation = doc_to_target(self.task_dict[task][split][doc_id])
#             visuals = [doc_to_visual(self.task_dict[task][split][doc_id])]
#             visuals = self.flatten(visuals)
#             image_sizes = [[visual.size[0], visual.size[1]] for visual in visuals]
#             if visuals:
#                 image = process_images(visuals, self._image_processor, self._config)
#                 if type(image) is list:
#                     image = [_image.to(dtype=torch.float16, device=self.device) for _image in image]
#                 else:
#                     image = image.to(dtype=torch.float16, device=self.device)
#             else:
#                 image = None

#             prompts_input = contexts[0] if isinstance(contexts, list) else contexts

#             if image is not None and len(image) != 0 and self.image_token not in prompts_input:
#                 """
#                 Three scenarios:
#                 1. No image, and there for, no image token should be added.
#                 2. image token is already specified in the context, so we don't need to add it.
#                 3. image token is not specified in the context and there is image inputs, so we need to add it. In this case, we add the image token at the beginning of the context and add a new line.
#                 """
#                 image_tokens = [self.image_token] * len(visuals)
#                 image_tokens = " ".join(image_tokens)
#                 prompts_input = (
#                     image_tokens + "\n" + (contexts[0] if isinstance(contexts, list) else contexts)
#                 )

#             # This is much safer for llama3, as we now have some object type in it
#             if "llama_3" in self.conv_template:
#                 conv = copy.deepcopy(conv_templates[self.conv_template])
#             else:
#                 conv = conv_templates[self.conv_template].copy()
#             conv.append_message(conv.roles[0], prompts_input)
#             conv.append_message(conv.roles[1], None)
#             prompt = conv.get_prompt()
#             pad_token_id = (
#                 self.tokenizer.pad_token_id
#                 if self.tokenizer.pad_token_id is not None
#                 else self.tokenizer.eos_token_id
#             )
#             contxt_id = (
#                 tokenizer_image_token(
#                     prompt, self.tokenizer, self.image_token_index, return_tensors="pt"
#                 )
#                 .unsqueeze(0)
#                 .to(self.device)
#             )
#             # Add the answer of the second role
#             conv.messages[1][1] = continuation

#             prompt = conv.get_prompt()
#             input_ids = (
#                 tokenizer_image_token(
#                     prompt, self.tokenizer, self.image_token_index, return_tensors="pt"
#                 )
#                 .unsqueeze(0)
#                 .to(self.device)
#             )
#             labels = input_ids.clone()
#             # Context part no need to calculate for loss
#             labels[0, : contxt_id.shape[1]] = -100
#             with torch.inference_mode():
#                 outputs = self.model(
#                     input_ids=input_ids,
#                     labels=labels,
#                     images=image,
#                     use_cache=True,
#                     image_sizes=image_sizes,
#                 )
#             loss = outputs["loss"]
#             # loss = torch.exp(loss)
#             logits = outputs["logits"]
#             greedy_tokens = logits.argmax(dim=-1)
#             cont_toks = input_ids[:, contxt_id.shape[1] :]  # [1, seq]
#             greedy_tokens = greedy_tokens[:, contxt_id.shape[1] : input_ids.shape[1]]  # [1, seq]
#             max_equal = (greedy_tokens == cont_toks).all()
#             res.append((float(loss.item()), bool(max_equal)))
#             pbar.update(1)
#         pbar.close()
#         return res

#     def flatten(self, input):
#         if not input or any(i is None for i in input):
#             return []
#         new_list = []
#         for i in input:
#             if i:
#                 for j in i:
#                     new_list.append(j)
#         return new_list

#     def generate_until(self, requests: list[Instance]) -> list[str]:
#         res = []

#         def _collate(x):
#             # the negative sign on len(toks) sorts descending - this has a few advantages:
#             # - time estimates will always be over not underestimates, which is more useful for planning
#             # - to know the size of a batch when going through the list, you know the first one is always the batch
#             #   padded context length. this is useful to simplify the batching logic and more importantly to make
#             #   automatic adaptive batches much much easier to implement
#             # - any OOMs will happen right away rather than near the end
#             toks = self.tok_encode(x[0])
#             return -len(toks), x[0]

#         # we group requests by their generation_kwargs,
#         # so that we don't try to execute e.g. greedy sampling and temp=0.8 sampling
#         # in the same batch.
#         re_ords = utils.Collator([reg.args for reg in requests], _collate, grouping=True)
#         chunks = re_ords.get_batched(n=self.batch_size, batch_fn=None)
#         num_iters = (
#             len(requests) // self.batch_size
#             if len(requests) % self.batch_size == 0
#             else len(requests) // self.batch_size + 1
#         )
#         pbar = tqdm(total=num_iters, disable=(self.rank != 0), desc="Model Responding")
#         for chunk in chunks:
#             contexts, all_gen_kwargs, doc_to_visual, doc_id, task, split = zip(*chunk, strict=False)
#             task = task[0]
#             split = split[0]
#             batched_visuals = [
#                 doc_to_visual[0](self.task_dict[task][split][ids]) for ids in doc_id
#             ]  # [B, N]
#             flattened_visuals = self.flatten(batched_visuals)  # [B*N]
#             # we assume all gen kwargs in the batch are the same
#             # this is safe to assume because the `grouper` object ensures it.
#             gen_kwargs = all_gen_kwargs[0]

#             # Set default values for until and max_new_tokens
#             until = [self.tok_decode(self.eot_token_id)]

#             # Update values from gen_kwargs if present
#             if "until" in gen_kwargs:
#                 until = gen_kwargs.pop("until")
#                 if isinstance(until, str):
#                     until = [until]
#                 elif not isinstance(until, list):
#                     raise ValueError(
#                         f"Expected `gen_kwargs['until']` to be of type Union[str,list] but got {type(until)}"
#                     )

#             if (
#                 "image_aspect_ratio" in gen_kwargs.keys()
#                 and "image_aspect_ratio" not in self._config.__dict__
#             ):
#                 # here we should pop it out of gen_kwargs so that it doesn't get passed to the model for next step of generation
#                 self._config.image_aspect_ratio = gen_kwargs.pop("image_aspect_ratio")
#                 eval_logger.info(f"Setting image aspect ratio: {self._config.image_aspect_ratio}")
#             # encode, pad, and truncate contexts for this batch
#             if flattened_visuals:
#                 image_tensor = process_images(
#                     flattened_visuals, self._image_processor, self._config
#                 )
#                 if type(image_tensor) is list:
#                     image_tensor = [
#                         _image.to(dtype=torch.float16, device=self.device)
#                         for _image in image_tensor
#                     ]
#                 else:
#                     image_tensor = image_tensor.to(dtype=torch.float16, device=self.device)
#             else:
#                 image_tensor = None

#             # prompts_input = contexts[0]

#             question_input = []

#             for visual, context in zip(batched_visuals, contexts, strict=False):
#                 if (
#                     image_tensor is not None
#                     and len(image_tensor) != 0
#                     and self.image_token not in context
#                 ):
#                     """
#                     Three scenarios:
#                     1. No image, and there for, no image token should be added.
#                     2. image token is already specified in the context, so we don't need to add it.
#                     3. image token is not specified in the context and there is image inputs, so we need to add it. In this case, we add the image token at the beginning of the context and add a new line.
#                     """
#                     image_tokens = (
#                         [self.image_token] * len(visual)
#                         if isinstance(visual, list)
#                         else [self.image_token]
#                     )
#                     image_tokens = " ".join(image_tokens)
#                     question = image_tokens + "\n" + context
#                 else:
#                     question = context
#                 # This is much safer for llama3, as we now have some object type in it
#                 if "llama_3" in self.conv_template:
#                     conv = copy.deepcopy(conv_templates[self.conv_template])
#                 else:
#                     conv = conv_templates[self.conv_template].copy()
#                 conv.append_message(conv.roles[0], question)
#                 conv.append_message(conv.roles[1], None)
#                 prompt_question = conv.get_prompt()
#                 question_input.append(prompt_question)

#             # input_ids = tokenizer_image_token(prompt, self.tokenizer, IMAGE_TOKEN_INDEX, return_tensors="pt").unsqueeze(0).to(self.device)
#             # preconfigure gen_kwargs with defaults
#             gen_kwargs["image_sizes"] = [
#                 flattened_visuals[idx].size for idx in range(len(flattened_visuals))
#             ]
#             if "max_new_tokens" not in gen_kwargs:
#                 gen_kwargs["max_new_tokens"] = 1024
#             if "temperature" not in gen_kwargs:
#                 gen_kwargs["temperature"] = 0
#             if "top_p" not in gen_kwargs:
#                 gen_kwargs["top_p"] = None
#             if "num_beams" not in gen_kwargs:
#                 gen_kwargs["num_beams"] = 1

#             input_ids_list = [
#                 tokenizer_image_token(
#                     prompt, self.tokenizer, self.image_token_index, return_tensors="pt"
#                 )
#                 for prompt in question_input
#             ]
#             pad_token_ids = (
#                 self.tokenizer.pad_token_id
#                 if self.tokenizer.pad_token_id is not None
#                 else self.tokenizer.eos_token_id
#             )
#             input_ids = self.pad_sequence(
#                 input_ids_list, batch_first=True, padding_value=pad_token_ids
#             ).to(self.device)
#             attention_masks = input_ids.ne(pad_token_ids).to(self.device)
#             # These steps are not in LLaVA's original code, but are necessary for generation to work
#             # TODO: attention to this major generation step...
#             try:
#                 cont = self.model.generate(
#                     input_ids,
#                     attention_mask=attention_masks,
#                     pad_token_id=pad_token_ids,
#                     images=image_tensor,
#                     image_sizes=gen_kwargs["image_sizes"],
#                     do_sample=True if gen_kwargs["temperature"] > 0 else False,
#                     temperature=gen_kwargs["temperature"],
#                     top_p=gen_kwargs["top_p"],
#                     num_beams=gen_kwargs["num_beams"],
#                     max_new_tokens=gen_kwargs["max_new_tokens"],
#                     use_cache=self.use_cache,
#                 )
#                 text_outputs = self.tokenizer.batch_decode(cont, skip_special_tokens=True)
#             except Exception as e:
#                 raise e
#                 eval_logger.error(f"Error {e} in generating")
#                 cont = ""
#                 text_outputs = [""]

#             # cont_toks_list = cont.tolist()
#             # for cont_toks, context in zip(cont_toks_list, contexts):
#             # discard context + left-padding toks if using causal decoder-only LMM
#             # if self.truncate_context:
#             #     cont_toks = cont_toks[input_ids.shape[1] :]
#             # use secondary stop seqs to cut off should-have-been-stopped content post-hoc
#             # if self.truncate_context:
#             #     for term in until:
#             #         if len(term) > 0:
#             #             # ignore '' separator,
#             #             # for seq2seq case where self.tok_decode(self.eot_token_id) = ''
#             #             text_outputs = text_outputs.split(term)[0]
#             res.extend(text_outputs)
#             self.cache_hook.add_partial("generate_until", (context, gen_kwargs), text_outputs)
#             pbar.update(1)
#             # reorder this group of results back to original unsorted form
#         res = re_ords.get_original(res)

#         pbar.close()
#         return res

#     def generate_until_multi_round(self, requests) -> list[str]:
#         raise NotImplementedError("TODO: Implement multi-round generation for LLaVA")
